# Entries are dictionary objects with
# name: string, arbitrary
# category: string, will be used to match StagingTable objects as inputs
#       for integration
# ds_family_abbr: string, will be used to match
#       StagingTable.data_source.parent.abbreviation for inputs
# ds_parent_abbr: string, will be used to match
#       StagingTable.data_source.parent.abbreviation for inputs
# columns: list, each member must be dict with
#               field: string, will match column names for inputs
#                       TODO: extend this to support lists, for
#                               many-to-one situations, condensing cols
#               type: string, must be a PostgreSQL column data type

DEFAULT_DEFINITIONS = [
    {
        'name': 'SID Standard Universe',
        'category': 'CORE_PROCESSED',
        'ds_family_abbr': 'SID', 
        'ds_parent_abbr': 'HCUP', # the parent of the real Definition.data_source
        'columns': [
            {'field': 'KEY', 'type': 'BIGINT'},
            {'field': 'STATE', 'type': 'VARCHAR'},
            {'field': 'YEAR', 'type': 'INT'},
            {'field': 'VISITLINK', 'type': 'BIGINT'},
            {'field': 'ZIP', 'type': 'VARCHAR'},
            {'field': 'ZIP3', 'type': 'VARCHAR'},
            {'field': 'DAYSTOEVENT', 'type': 'INT'},
            {'field': 'RACE', 'type': 'INT'},
            {'field': 'FEMALE', 'type': 'INT'},
            {'field': 'AGE', 'type': 'INT'},
            {'field': 'AGEDAY', 'type': 'INT'},
            {'field': 'AGEMONTH', 'type': 'INT'},
            {'field': 'PAY1', 'type': 'INT'},
            {'field': 'PAY2', 'type': 'INT'},
            {'field': 'DISPUB04', 'type': 'INT'},
            {'field': 'DISPUB92', 'type': 'INT'},
            {'field': 'ASOURCE', 'type': 'INT'},
            {'field': 'AMONTH', 'type': 'INT'},
            {'field': 'DRG', 'type': 'VARCHAR'},
            {'field': 'DRGVER', 'type': 'VARCHAR'},
            {'field': 'TRAN_IN', 'type': 'INT'},
            {'field': 'TRAN_OUT', 'type': 'INT'},
            {'field': 'DIED', 'type': 'INT'},
            {'field': 'LOS', 'type': 'INT'},
            {'field': 'ahaid', 'type': 'VARCHAR'},
            {'field': 'hospid', 'type': 'INT'},
            {'field': 'hospstco', 'type': 'INT'},
            {'field': 'cm_aids', 'type': 'INT'},
            {'field': 'cm_alcohol', 'type': 'INT'},
            {'field': 'cm_anemdef', 'type': 'INT'},
            {'field': 'cm_arth', 'type': 'INT'},
            {'field': 'cm_bldloss', 'type': 'INT'},
            {'field': 'cm_chf', 'type': 'INT'},
            {'field': 'cm_chrnlung', 'type': 'INT'},
            {'field': 'cm_coag', 'type': 'INT'},
            {'field': 'cm_depress', 'type': 'INT'},
            {'field': 'cm_dm', 'type': 'INT'},
            {'field': 'cm_dmcx', 'type': 'INT'},
            {'field': 'cm_drug', 'type': 'INT'},
            {'field': 'cm_htn_c', 'type': 'INT'},
            {'field': 'cm_hypothy', 'type': 'INT'},
            {'field': 'cm_liver', 'type': 'INT'},
            {'field': 'cm_lymph', 'type': 'INT'},
            {'field': 'cm_lytes', 'type': 'INT'},
            {'field': 'cm_mets', 'type': 'INT'},
            {'field': 'cm_neuro', 'type': 'INT'},
            {'field': 'cm_obese', 'type': 'INT'},
            {'field': 'cm_para', 'type': 'INT'},
            {'field': 'cm_perivasc', 'type': 'INT'},
            {'field': 'cm_psych', 'type': 'INT'},
            {'field': 'cm_pulmcirc', 'type': 'INT'},
            {'field': 'cm_renlfail', 'type': 'INT'},
            {'field': 'cm_tumor', 'type': 'INT'},
            {'field': 'cm_ulcer', 'type': 'INT'},
            {'field': 'cm_valve', 'type': 'INT'},
            {'field': 'cm_wghtloss', 'type': 'INT'},
            {'field': 'chron1', 'type': 'INT'},
            {'field': 'chron2', 'type': 'INT'},
            {'field': 'chron3', 'type': 'INT'},
            {'field': 'chron4', 'type': 'INT'},
            {'field': 'chron5', 'type': 'INT'},
            {'field': 'chron6', 'type': 'INT'},
            {'field': 'chron7', 'type': 'INT'},
            {'field': 'chron8', 'type': 'INT'},
            {'field': 'chron9', 'type': 'INT'},
            {'field': 'chron10', 'type': 'INT'},
            {'field': 'chron11', 'type': 'INT'},
            {'field': 'chron12', 'type': 'INT'},
            {'field': 'chron13', 'type': 'INT'},
            {'field': 'chron14', 'type': 'INT'},
            {'field': 'chron15', 'type': 'INT'},
            {'field': 'chron16', 'type': 'INT'},
            {'field': 'chron17', 'type': 'INT'},
            {'field': 'chron18', 'type': 'INT'},
            {'field': 'chron19', 'type': 'INT'},
            {'field': 'chron20', 'type': 'INT'},
            {'field': 'chron21', 'type': 'INT'},
            {'field': 'chron22', 'type': 'INT'},
            {'field': 'chron23', 'type': 'INT'},
            {'field': 'chron24', 'type': 'INT'},
            {'field': 'chron25', 'type': 'INT'},
            {'field': 'chronb1', 'type': 'INT'},
            {'field': 'chronb2', 'type': 'INT'},
            {'field': 'chronb3', 'type': 'INT'},
            {'field': 'chronb4', 'type': 'INT'},
            {'field': 'chronb5', 'type': 'INT'},
            {'field': 'chronb6', 'type': 'INT'},
            {'field': 'chronb7', 'type': 'INT'},
            {'field': 'chronb8', 'type': 'INT'},
            {'field': 'chronb9', 'type': 'INT'},
            {'field': 'chronb10', 'type': 'INT'},
            {'field': 'chronb11', 'type': 'INT'},
            {'field': 'chronb12', 'type': 'INT'},
            {'field': 'chronb13', 'type': 'INT'},
            {'field': 'chronb14', 'type': 'INT'},
            {'field': 'chronb15', 'type': 'INT'},
            {'field': 'chronb16', 'type': 'INT'},
            {'field': 'chronb17', 'type': 'INT'},
            {'field': 'chronb18', 'type': 'INT'},
            {'field': 'chronb19', 'type': 'INT'},
            {'field': 'chronb20', 'type': 'INT'},
            {'field': 'chronb21', 'type': 'INT'},
            {'field': 'chronb22', 'type': 'INT'},
            {'field': 'chronb23', 'type': 'INT'},
            {'field': 'chronb24', 'type': 'INT'},
            {'field': 'chronb25', 'type': 'INT'},
            {'field': 'dxmccs1', 'type': 'VARCHAR'},
            {'field': 'dxmccs2', 'type': 'VARCHAR'},
            {'field': 'dxmccs3', 'type': 'VARCHAR'},
            {'field': 'dxmccs4', 'type': 'VARCHAR'},
            {'field': 'dxmccs5', 'type': 'VARCHAR'},
            {'field': 'dxmccs6', 'type': 'VARCHAR'},
            {'field': 'dxmccs7', 'type': 'VARCHAR'},
            {'field': 'dxmccs8', 'type': 'VARCHAR'},
            {'field': 'dxmccs9', 'type': 'VARCHAR'},
            {'field': 'dxmccs10', 'type': 'VARCHAR'},
            {'field': 'dxmccs11', 'type': 'VARCHAR'},
            {'field': 'dxmccs12', 'type': 'VARCHAR'},
            {'field': 'dxmccs13', 'type': 'VARCHAR'},
            {'field': 'dxmccs14', 'type': 'VARCHAR'},
            {'field': 'dxmccs15', 'type': 'VARCHAR'},
            {'field': 'dxmccs16', 'type': 'VARCHAR'},
            {'field': 'dxmccs17', 'type': 'VARCHAR'},
            {'field': 'dxmccs18', 'type': 'VARCHAR'},
            {'field': 'dxmccs19', 'type': 'VARCHAR'},
            {'field': 'dxmccs20', 'type': 'VARCHAR'},
            {'field': 'dxmccs21', 'type': 'VARCHAR'},
            {'field': 'dxmccs22', 'type': 'VARCHAR'},
            {'field': 'dxmccs23', 'type': 'VARCHAR'},
            {'field': 'dxmccs24', 'type': 'VARCHAR'},
            {'field': 'dxmccs25', 'type': 'VARCHAR'},
            {'field': 'ccsmgn1', 'type': 'INT'},
            {'field': 'ccsmgn2', 'type': 'INT'},
            {'field': 'ccsmgn3', 'type': 'INT'},
            {'field': 'ccsmgn4', 'type': 'INT'},
            {'field': 'ccsmgn5', 'type': 'INT'},
            {'field': 'ccsmgn6', 'type': 'INT'},
            {'field': 'ccsmgn7', 'type': 'INT'},
            {'field': 'ccsmgn8', 'type': 'INT'},
            {'field': 'ccsmgn9', 'type': 'INT'},
            {'field': 'ccsmsp1', 'type': 'INT'},
            {'field': 'ccsmsp2', 'type': 'INT'},
            {'field': 'ccsmsp3', 'type': 'INT'},
            {'field': 'ccsmsp4', 'type': 'INT'},
            {'field': 'ccsmsp5', 'type': 'INT'},
            {'field': 'ccsmsp6', 'type': 'INT'},
            {'field': 'ccsmsp7', 'type': 'INT'},
            {'field': 'ccsmsp8', 'type': 'INT'},
            {'field': 'ccsmsp9', 'type': 'INT'},
            {'field': 'eccsmgn1', 'type': 'INT'},
            {'field': 'eccsmgn2', 'type': 'INT'},
            {'field': 'eccsmgn3', 'type': 'INT'},
            {'field': 'eccsmgn4', 'type': 'INT'},
            {'field': 'eccsmgn5', 'type': 'INT'},
            {'field': 'e_mccs1', 'type': 'VARCHAR'},
            {'field': 'e_mccs2', 'type': 'VARCHAR'},
            {'field': 'e_mccs3', 'type': 'VARCHAR'},
            {'field': 'e_mccs4', 'type': 'VARCHAR'},
            {'field': 'e_mccs5', 'type': 'VARCHAR'},
            {'field': 'e_mccs6', 'type': 'VARCHAR'},
            {'field': 'pclass1', 'type': 'INT'},
            {'field': 'pclass2', 'type': 'INT'},
            {'field': 'pclass3', 'type': 'INT'},
            {'field': 'pclass4', 'type': 'INT'},
            {'field': 'pclass5', 'type': 'INT'},
            {'field': 'pclass6', 'type': 'INT'},
            {'field': 'pclass7', 'type': 'INT'},
            {'field': 'pclass8', 'type': 'INT'},
            {'field': 'pclass9', 'type': 'INT'},
            {'field': 'pclass10', 'type': 'INT'},
            {'field': 'pclass11', 'type': 'INT'},
            {'field': 'pclass12', 'type': 'INT'},
            {'field': 'u_icu', 'type': 'INT'},
            {'field': 'u_ccu', 'type': 'INT'},
            {'field': 'u_newbn2l', 'type': 'INT'},
            {'field': 'u_newbn3l', 'type': 'INT'},
            {'field': 'u_newbn4l', 'type': 'INT'},
            {'field': 'u_cath', 'type': 'INT'},
            {'field': 'u_stress', 'type': 'INT'},
            {'field': 'u_echo', 'type': 'INT'},
            {'field': 'u_ekg', 'type': 'INT'},
            {'field': 'u_ctscan', 'type': 'INT'},
            {'field': 'u_chestxray', 'type': 'INT'},
            {'field': 'u_ultrasound', 'type': 'INT'},
            {'field': 'u_eeg', 'type': 'INT'},
            {'field': 'u_nucmed', 'type': 'INT'},
            {'field': 'u_mrt', 'type': 'INT'},
            {'field': 'u_pacemaker', 'type': 'INT'},
            {'field': 'u_othimplants', 'type': 'INT'},
            {'field': 'u_lithotripsy', 'type': 'INT'},
            {'field': 'u_occtherapy', 'type': 'INT'},
            {'field': 'u_phytherapy', 'type': 'INT'},
            {'field': 'u_radtherapy', 'type': 'INT'},
            {'field': 'u_resptherapy', 'type': 'INT'},
            {'field': 'u_speechtherapy', 'type': 'INT'},
            {'field': 'u_blood', 'type': 'INT'},
            {'field': 'u_dialysis', 'type': 'INT'},
            {'field': 'u_epo', 'type': 'INT'},
            {'field': 'u_mhsa', 'type': 'INT'},
            {'field': 'u_ed', 'type': 'INT'},
            {'field': 'u_observation', 'type': 'INT'},
            {'field': 'u_organacq', 'type': 'INT'},
            {'field': 'pr1', 'type': 'VARCHAR'},
            {'field': 'pr2', 'type': 'VARCHAR'},
            {'field': 'pr3', 'type': 'VARCHAR'},
            {'field': 'pr4', 'type': 'VARCHAR'},
            {'field': 'pr5', 'type': 'VARCHAR'},
            {'field': 'pr6', 'type': 'VARCHAR'},
            {'field': 'pr7', 'type': 'VARCHAR'},
            {'field': 'pr8', 'type': 'VARCHAR'},
            {'field': 'pr9', 'type': 'VARCHAR'},
            {'field': 'pr10', 'type': 'VARCHAR'},
            {'field': 'pr11', 'type': 'VARCHAR'},
            {'field': 'pr12', 'type': 'VARCHAR'},
            {'field': 'pr13', 'type': 'VARCHAR'},
            {'field': 'pr14', 'type': 'VARCHAR'},
            {'field': 'pr15', 'type': 'VARCHAR'},
            {'field': 'pr16', 'type': 'VARCHAR'},
            {'field': 'pr17', 'type': 'VARCHAR'},
            {'field': 'pr18', 'type': 'VARCHAR'},
            {'field': 'pr19', 'type': 'VARCHAR'},
            {'field': 'pr20', 'type': 'VARCHAR'},
            {'field': 'pr21', 'type': 'VARCHAR'},
            {'field': 'pr22', 'type': 'VARCHAR'},
            {'field': 'pr23', 'type': 'VARCHAR'},
            {'field': 'pr24', 'type': 'VARCHAR'},
            {'field': 'pr25', 'type': 'VARCHAR'},
            {'field': 'pryear1', 'type': 'INT'},
            {'field': 'pryear2', 'type': 'INT'},
            {'field': 'pryear3', 'type': 'INT'},
            {'field': 'pryear4', 'type': 'INT'},
            {'field': 'pryear5', 'type': 'INT'},
            {'field': 'pryear6', 'type': 'INT'},
            {'field': 'pryear7', 'type': 'INT'},
            {'field': 'pryear8', 'type': 'INT'},
            {'field': 'pryear9', 'type': 'INT'},
            {'field': 'pryear10', 'type': 'INT'},
            {'field': 'pryear11', 'type': 'INT'},
            {'field': 'pryear12', 'type': 'INT'},
            {'field': 'pryear13', 'type': 'INT'},
            {'field': 'pryear14', 'type': 'INT'},
            {'field': 'pryear15', 'type': 'INT'},
            {'field': 'pryear16', 'type': 'INT'},
            {'field': 'pryear17', 'type': 'INT'},
            {'field': 'pryear18', 'type': 'INT'},
            {'field': 'pryear19', 'type': 'INT'},
            {'field': 'pryear20', 'type': 'INT'},
            {'field': 'pryear21', 'type': 'INT'},
            {'field': 'pryear22', 'type': 'INT'},
            {'field': 'pryear23', 'type': 'INT'},
            {'field': 'pryear24', 'type': 'INT'},
            {'field': 'pryear25', 'type': 'INT'},
            {'field': 'prmonth1', 'type': 'INT'},
            {'field': 'prmonth2', 'type': 'INT'},
            {'field': 'prmonth3', 'type': 'INT'},
            {'field': 'prmonth4', 'type': 'INT'},
            {'field': 'prmonth5', 'type': 'INT'},
            {'field': 'prmonth6', 'type': 'INT'},
            {'field': 'prmonth7', 'type': 'INT'},
            {'field': 'prmonth8', 'type': 'INT'},
            {'field': 'prmonth9', 'type': 'INT'},
            {'field': 'prmonth10', 'type': 'INT'},
            {'field': 'prmonth11', 'type': 'INT'},
            {'field': 'prmonth12', 'type': 'INT'},
            {'field': 'prmonth13', 'type': 'INT'},
            {'field': 'prmonth14', 'type': 'INT'},
            {'field': 'prmonth15', 'type': 'INT'},
            {'field': 'prmonth16', 'type': 'INT'},
            {'field': 'prmonth17', 'type': 'INT'},
            {'field': 'prmonth18', 'type': 'INT'},
            {'field': 'prmonth19', 'type': 'INT'},
            {'field': 'prmonth20', 'type': 'INT'},
            {'field': 'prmonth21', 'type': 'INT'},
            {'field': 'prmonth22', 'type': 'INT'},
            {'field': 'prmonth23', 'type': 'INT'},
            {'field': 'prmonth24', 'type': 'INT'},
            {'field': 'prmonth25', 'type': 'INT'},
            {'field': 'prday1', 'type': 'INT'},
            {'field': 'prday2', 'type': 'INT'},
            {'field': 'prday3', 'type': 'INT'},
            {'field': 'prday4', 'type': 'INT'},
            {'field': 'prday5', 'type': 'INT'},
            {'field': 'prday6', 'type': 'INT'},
            {'field': 'prday7', 'type': 'INT'},
            {'field': 'prday8', 'type': 'INT'},
            {'field': 'prday9', 'type': 'INT'},
            {'field': 'prday10', 'type': 'INT'},
            {'field': 'prday11', 'type': 'INT'},
            {'field': 'prday12', 'type': 'INT'},
            {'field': 'prday13', 'type': 'INT'},
            {'field': 'prday14', 'type': 'INT'},
            {'field': 'prday15', 'type': 'INT'},
            {'field': 'prday16', 'type': 'INT'},
            {'field': 'prday17', 'type': 'INT'},
            {'field': 'prday18', 'type': 'INT'},
            {'field': 'prday19', 'type': 'INT'},
            {'field': 'prday20', 'type': 'INT'},
            {'field': 'prday21', 'type': 'INT'},
            {'field': 'prday22', 'type': 'INT'},
            {'field': 'prday23', 'type': 'INT'},
            {'field': 'prday24', 'type': 'INT'},
            {'field': 'prday25', 'type': 'INT'},
            {'field': 'prccs1', 'type': 'VARCHAR'},
            {'field': 'prccs2', 'type': 'VARCHAR'},
            {'field': 'prccs3', 'type': 'VARCHAR'},
            {'field': 'prccs4', 'type': 'VARCHAR'},
            {'field': 'prccs5', 'type': 'VARCHAR'},
            {'field': 'prccs6', 'type': 'VARCHAR'},
            {'field': 'prccs7', 'type': 'VARCHAR'},
            {'field': 'prccs8', 'type': 'VARCHAR'},
            {'field': 'prccs9', 'type': 'VARCHAR'},
            {'field': 'prccs10', 'type': 'VARCHAR'},
            {'field': 'prccs11', 'type': 'VARCHAR'},
            {'field': 'prccs12', 'type': 'VARCHAR'},
            {'field': 'prccs13', 'type': 'VARCHAR'},
            {'field': 'prccs14', 'type': 'VARCHAR'},
            {'field': 'prccs15', 'type': 'VARCHAR'},
            {'field': 'prccs16', 'type': 'VARCHAR'},
            {'field': 'prccs17', 'type': 'VARCHAR'},
            {'field': 'prccs18', 'type': 'VARCHAR'},
            {'field': 'prccs19', 'type': 'VARCHAR'},
            {'field': 'prccs20', 'type': 'VARCHAR'},
            {'field': 'prccs21', 'type': 'VARCHAR'},
            {'field': 'prccs22', 'type': 'VARCHAR'},
            {'field': 'prccs23', 'type': 'VARCHAR'},
            {'field': 'prccs24', 'type': 'VARCHAR'},
            {'field': 'prccs25', 'type': 'VARCHAR'},
            {'field': 'dx1', 'type': 'VARCHAR'},
            {'field': 'dx2', 'type': 'VARCHAR'},
            {'field': 'dx3', 'type': 'VARCHAR'},
            {'field': 'dx4', 'type': 'VARCHAR'},
            {'field': 'dx5', 'type': 'VARCHAR'},
            {'field': 'dx6', 'type': 'VARCHAR'},
            {'field': 'dx7', 'type': 'VARCHAR'},
            {'field': 'dx8', 'type': 'VARCHAR'},
            {'field': 'dx9', 'type': 'VARCHAR'},
            {'field': 'dx10', 'type': 'VARCHAR'},
            {'field': 'dx11', 'type': 'VARCHAR'},
            {'field': 'dx12', 'type': 'VARCHAR'},
            {'field': 'dx13', 'type': 'VARCHAR'},
            {'field': 'dx14', 'type': 'VARCHAR'},
            {'field': 'dx15', 'type': 'VARCHAR'},
            {'field': 'dx16', 'type': 'VARCHAR'},
            {'field': 'dx17', 'type': 'VARCHAR'},
            {'field': 'dx18', 'type': 'VARCHAR'},
            {'field': 'dx19', 'type': 'VARCHAR'},
            {'field': 'dx20', 'type': 'VARCHAR'},
            {'field': 'dx21', 'type': 'VARCHAR'},
            {'field': 'dx22', 'type': 'VARCHAR'},
            {'field': 'dx23', 'type': 'VARCHAR'},
            {'field': 'dx24', 'type': 'VARCHAR'},
            {'field': 'dx25', 'type': 'VARCHAR'},
            {'field': 'dxpoa1', 'type': 'VARCHAR'},
            {'field': 'dxpoa2', 'type': 'VARCHAR'},
            {'field': 'dxpoa3', 'type': 'VARCHAR'},
            {'field': 'dxpoa4', 'type': 'VARCHAR'},
            {'field': 'dxpoa5', 'type': 'VARCHAR'},
            {'field': 'dxpoa6', 'type': 'VARCHAR'},
            {'field': 'dxpoa7', 'type': 'VARCHAR'},
            {'field': 'dxpoa8', 'type': 'VARCHAR'},
            {'field': 'dxpoa9', 'type': 'VARCHAR'},
            {'field': 'dxpoa10', 'type': 'VARCHAR'},
            {'field': 'dxpoa11', 'type': 'VARCHAR'},
            {'field': 'dxpoa12', 'type': 'VARCHAR'},
            {'field': 'dxpoa13', 'type': 'VARCHAR'},
            {'field': 'dxpoa14', 'type': 'VARCHAR'},
            {'field': 'dxpoa15', 'type': 'VARCHAR'},
            {'field': 'dxpoa16', 'type': 'VARCHAR'},
            {'field': 'dxpoa17', 'type': 'VARCHAR'},
            {'field': 'dxpoa18', 'type': 'VARCHAR'},
            {'field': 'dxpoa19', 'type': 'VARCHAR'},
            {'field': 'dxpoa20', 'type': 'VARCHAR'},
            {'field': 'dxpoa21', 'type': 'VARCHAR'},
            {'field': 'dxpoa22', 'type': 'VARCHAR'},
            {'field': 'dxpoa23', 'type': 'VARCHAR'},
            {'field': 'dxpoa24', 'type': 'VARCHAR'},
            {'field': 'dxpoa25', 'type': 'VARCHAR'},
            {'field': 'dxccs1', 'type': 'VARCHAR'},
            {'field': 'dxccs2', 'type': 'VARCHAR'},
            {'field': 'dxccs3', 'type': 'VARCHAR'},
            {'field': 'dxccs4', 'type': 'VARCHAR'},
            {'field': 'dxccs5', 'type': 'VARCHAR'},
            {'field': 'dxccs6', 'type': 'VARCHAR'},
            {'field': 'dxccs7', 'type': 'VARCHAR'},
            {'field': 'dxccs8', 'type': 'VARCHAR'},
            {'field': 'dxccs9', 'type': 'VARCHAR'},
            {'field': 'dxccs10', 'type': 'VARCHAR'},
            {'field': 'dxccs11', 'type': 'VARCHAR'},
            {'field': 'dxccs12', 'type': 'VARCHAR'},
            {'field': 'dxccs13', 'type': 'VARCHAR'},
            {'field': 'dxccs14', 'type': 'VARCHAR'},
            {'field': 'dxccs15', 'type': 'VARCHAR'},
            {'field': 'dxccs16', 'type': 'VARCHAR'},
            {'field': 'dxccs17', 'type': 'VARCHAR'},
            {'field': 'dxccs18', 'type': 'VARCHAR'},
            {'field': 'dxccs19', 'type': 'VARCHAR'},
            {'field': 'dxccs20', 'type': 'VARCHAR'},
            {'field': 'dxccs21', 'type': 'VARCHAR'},
            {'field': 'dxccs22', 'type': 'VARCHAR'},
            {'field': 'dxccs23', 'type': 'VARCHAR'},
            {'field': 'dxccs24', 'type': 'VARCHAR'},
            {'field': 'dxccs25', 'type': 'VARCHAR'},
            {'field': 'prmccs1', 'type': 'VARCHAR'},
            {'field': 'prmccs2', 'type': 'VARCHAR'},
            {'field': 'prmccs3', 'type': 'VARCHAR'},
            {'field': 'prmccs4', 'type': 'VARCHAR'},
            {'field': 'prmccs5', 'type': 'VARCHAR'},
            {'field': 'prmccs6', 'type': 'VARCHAR'},
            {'field': 'prmccs7', 'type': 'VARCHAR'},
            {'field': 'prmccs8', 'type': 'VARCHAR'},
            {'field': 'prmccs9', 'type': 'VARCHAR'},
            {'field': 'prmccs10', 'type': 'VARCHAR'},
            {'field': 'prmccs11', 'type': 'VARCHAR'},
            {'field': 'prmccs12', 'type': 'VARCHAR'},
            {'field': 'prmccs13', 'type': 'VARCHAR'},
            {'field': 'prmccs14', 'type': 'VARCHAR'},
            {'field': 'prmccs15', 'type': 'VARCHAR'},
            {'field': 'prmccs16', 'type': 'VARCHAR'},
            {'field': 'prmccs17', 'type': 'VARCHAR'},
            {'field': 'prmccs18', 'type': 'VARCHAR'},
            {'field': 'prmccs19', 'type': 'VARCHAR'},
            {'field': 'prmccs20', 'type': 'VARCHAR'},
            {'field': 'prmccs21', 'type': 'VARCHAR'},
            {'field': 'prmccs22', 'type': 'VARCHAR'},
            {'field': 'prmccs23', 'type': 'VARCHAR'},
            {'field': 'prmccs24', 'type': 'VARCHAR'},
            {'field': 'prmccs25', 'type': 'VARCHAR'}
        ]
    },
    # Long table (lookup table) definitions follow.    
    # If I had more time, these would be drawn from PyHCUP.
    # But their definition in PyHCUP would require a non-
    # trivial amount of reworking to get into a portable
    # format.
    # Alas!
    {
        'name': 'SID Charges (long)',
        'category': 'CHARGES',
        'ds_family_abbr': 'SID', 
        'ds_parent_abbr': 'HCUP', # the parent of the real Definition.data_source
        'columns': [
            {'field': 'KEY', 'type': 'BIGINT'},
            {'field': 'VISITLINK', 'type': 'BIGINT'},
            {'field': 'DAYSTOEVENT', 'type': 'INT'},
            {'field': 'YEAR', 'type': 'INT'},
            {'field': 'STATE', 'type': 'VARCHAR'},
            {'field': 'UNITS', 'type': 'INT'},
            {'field': 'REVCODE', 'type': 'VARCHAR'},
            {'field': 'RATE', 'type': 'NUMERIC(16, 2)'},
            {'field': 'CHARGE', 'type': 'NUMERIC(16, 2)'},
            {'field': 'CPTHCPCS', 'type': 'VARCHAR'},
            {'field': 'CPTMOD1', 'type': 'VARCHAR'},
            {'field': 'CPTMOD2', 'type': 'VARCHAR'},
            {'field': 'GROUP_NUMBER', 'type': 'INT'}
        ]
    },
    {
        'name': 'SID Diagnoses (long)',
        'category': 'DX',
        'ds_family_abbr': 'SID', 
        'ds_parent_abbr': 'HCUP', # the parent of the real Definition.data_source
        'columns': [
            {'field': 'KEY', 'type': 'BIGINT'},
            {'field': 'VISITLINK', 'type': 'BIGINT'},
            {'field': 'DAYSTOEVENT', 'type': 'INT'},
            {'field': 'YEAR', 'type': 'INT'},
            {'field': 'STATE', 'type': 'VARCHAR'},
            {'field': 'DX', 'type': 'VARCHAR'},
            {'field': 'DXV', 'type': 'VARCHAR'},
            {'field': 'DXCCS', 'type': 'VARCHAR'},
            {'field': 'DXPOA', 'type': 'VARCHAR'},
            {'field': 'DXatAdmit', 'type': 'VARCHAR'},
            {'field': 'TMDX', 'type': 'VARCHAR'},
            {'field': 'GROUP_NUMBER', 'type': 'INT'}
        ]
    },
    {
        'name': 'SID Procedures (long)',
        'category': 'PR',
        'ds_family_abbr': 'SID', 
        'ds_parent_abbr': 'HCUP', # the parent of the real Definition.data_source
        'columns': [
            {'field': 'KEY', 'type': 'BIGINT'},
            {'field': 'VISITLINK', 'type': 'BIGINT'},
            {'field': 'DAYSTOEVENT', 'type': 'INT'},
            {'field': 'YEAR', 'type': 'INT'},
            {'field': 'STATE', 'type': 'VARCHAR'},
            {'field': 'PR', 'type': 'VARCHAR'},
            {'field': 'PRCCS', 'type': 'VARCHAR'},
            {'field': 'PRDATE', 'type': 'VARCHAR'},
            {'field': 'PRDAY', 'type': 'INT'},
            {'field': 'PRMONTH', 'type': 'INT'},
            {'field': 'PRYEAR', 'type': 'INT'},
            {'field': 'PRV', 'type': 'VARCHAR'},
            {'field': 'PCLASS', 'type': 'VARCHAR'},
            {'field': 'PRMCCS', 'type': 'VARCHAR'},
            {'field': 'GROUP_NUMBER', 'type': 'INT'}
        ]
    },
    {
        'name': 'SID Utilization Flags (long)',
        'category': 'UFLAGS',
        'ds_family_abbr': 'SID', 
        'ds_parent_abbr': 'HCUP', # the parent of the real Definition.data_source
        'columns': [
            {'field': 'KEY', 'type': 'BIGINT'},
            {'field': 'VISITLINK', 'type': 'BIGINT'},
            {'field': 'DAYSTOEVENT', 'type': 'INT'},
            {'field': 'YEAR', 'type': 'INT'},
            {'field': 'STATE', 'type': 'VARCHAR'},
            {'field': 'NAME', 'type': 'VARCHAR'},
            {'field': 'VALUE', 'type': 'INT'}
        ]
    }
]



# core Python packages
import datetime
import itertools
import logging
import os



# third party packages
import pyhcup


# django packages
from django.utils.timezone import utc
from django.conf import settings
#import shortcuts to use with raw sql (!) later
from django.core.exceptions import ObjectDoesNotExist


# local project and app packages
from djhcup_integration.models import Edition, Definition, Column, ColumnName, IntegrationTable, IntTblStgTblMembers, MappedColumn


# sister app stuff
from djhcup_staging.models import StagingTable
from djhcup_staging.utils.misc import get_pgcnxn


# start a logger
logger = logging.getLogger('default')


# make these tasks app-agnostic with @celery.shared_task
import celery
from celery.result import AsyncResult

@celery.shared_task
def generate_tables(e_pk, d_pk_lst):
    """
    This is the task to call for creating a new set of IntegrationTable
    objects and their corresponding database tables.
    
    Parameters:
    e_pk: integer (required)
    Primary key for integration Edition model object.
    
    Create a new Edition object and pass its primary key as edition_pk.
    
    d_lst: list (required)
    List of primary keys for integration Definition model objects.
    
    In most cases, pass something equivalent to 
    [x.pk for x in Definition.objects.all()]
    """
    integrations = celery.group(integrate_tables.s(d_pk, e_pk) for d_pk in d_pk_lst)
    """
    EDITION_CATS = [
        # a list of categories to include in each edition of an integration table set
        dict(val='CORE_PROCESSED', long_entry=None, ds_abbr='CORE'),
        dict(val='DX', long_entry='DX'),
        dict(val='PR', long_entry='PR'),
        dict(val='CHARGES', long_entry='CHGS'),
        dict(val='UFLAGS', long_entry='UFLAGS'),
    ]
    
    # for each category of data
        # grab StagingTables with that category
        # for each unique DataSource
            # filter StagingTables down to those matching DataSource (within category)
            
    # need to somehow deal with extra columns--might just have to define integration/reporting field sets or something
    create_sql = pyhcup.db.long_table_sql(db_table, str(source.abbreviation))
    index_fields = ['KEY', 'VISITLINK', 'STATE', 'YEAR']
    """
    return integrations()



@celery.shared_task
def integrate_stg_table(stgtbl_in_pk, int_tbl_pk, def_pk, dry_run=False):
    """
    Integrates StagingTable object with pk stgtbl_in_pk
    into IntegrationTable with pk int_tbl_pk
    according to Definition with pk def_pk.
    
    If dry_run == True, no changes will be made to the database.
    """
    
    # initiate a connection
    cnxn = get_pgcnxn()
    
    # find corresponding objects
    try:
        sttbl = StagingTable.objects.get(pk=stgtbl_in_pk)
        int_tbl = IntegrationTable.objects.get(pk=int_tbl_pk)
        d = Definition.objects.get(pk=def_pk)
    except ObjectDoesNotExist:
        raise ObjectDoesNotExist("Unable to find one of StagingTable {s}, IntegrationTable {i}, or Definition {d}" \
            .format(s=stgtbl_in_pk, i=int_tbl_pk, d=def_pk))
    
    # isolate state and year
    state = sttbl.state
    year = sttbl.year
    
    # create a new integration member, which holds staging tables pushed into int table
    if not dry_run:
        int_member = IntTblStgTblMembers(
            integrationtable = int_tbl,
            stagingtable = sttbl,
        )
        int_member.save()
    
    # instantiate lists for fields in and out
    fields_in_lst = []
    fields_out_lst = []
    
    # grab columns present in staging table
    st_cols = [s.lower() for s in pyhcup.db.pg_getcols(cnxn, sttbl.name)]
    
    for c in st_cols:
        matches = d.columns.filter(fields_in__value__iexact=c)
        if len(matches) == 1:
            matched_col = matches[0]
            matched_cn = ColumnName.objects.filter(fields_in=matched_col, value__iexact=c)[0]
            
            if not dry_run:
                mapped_col = MappedColumn(
                    inttblstgtblmembers=int_member,
                    column=matched_col,
                    matched_in=matched_cn)
                mapped_col.save()
            
            fields_in_lst.append(c)
            fields_out_lst.append(matched_cn.value)
        else:
            logger.debug(
                "Unable to resolve {sttbl}.{c}; skipping".format(
                    sttbl=sttbl.name,
                    c=c
                )
            )
    
    # shovel the resolved columns from source to integration table
    if not dry_run:
        shoveled = pyhcup.db.pg_shovel(cnxn, tbl_source=sttbl.name,
                                       tbl_destination=int_tbl.name,
                                       fields_in=fields_in_lst,
                                       fields_out=fields_out_lst)
        int_tbl.wip = False
        int_tbl.save()
    else:
        shoveled = 0
    
    logger.info("Shoveled {ct} rows from StagingTable {st_pk} ({tbl})" \
        .format(ct=shoveled, st_pk=sttbl.pk, tbl=sttbl.name)
    )
    
    logger.debug(
        "TABLE: {tbl}\nROWS: {rows}\nFIELDS IN: {fields_in}\nFIELDS OUT: {fields_out}" \
            .format(
                tbl=sttbl,
                rows=shoveled,
                fields_in=fields_in_lst,
                fields_out=fields_out_lst
            )
    )
    
    return dict(
        state=state,
        year=year,
        shoveled=shoveled,
        StagingTable=sttbl,
        fields_in=fields_in_lst,
        fields_out=fields_out_lst
    )


@celery.shared_task
def integrate_tables(definition_pk, edition_pk, indexes=['key', 'state', 'year', 'visitlink']):
    """
    Creates a new table integration using provided definition and edition.
    
    Dispatches individual table imports in parallel.
    
    TODO: Consider how much, if any, of these things should be preserved
    as provenance somewhere, somehow. Is description field appropriate?
    Alternatively, maybe a logging filter could be used, or an
    additional handler, that sends certain provenance items away for
    safekeeping.
    """
    logger.debug("integrate_tables() called with definition_pk {d} and edition_pk {e}" \
        .format(d=definition_pk, e=edition_pk)
    )
    
    try:
        # grab the definition for which we'll make an integration table
        d = Definition.objects.get(pk=definition_pk)
    except ObjectDoesNotExist:
        raise ObjectDoesNotExist("Function integrate_tables() was unable to find Definition with primary key %s" % definition_pk)
    
    try:
        e = Edition.objects.get(pk=edition_pk)
    except ObjectDoesNotExist:
        raise ObjectDoesNotExist("Function integrate_tables() was unable to find Edition with primary key %s" % edition_pk)
    
    cnxn = get_pgcnxn()

    # use the Edition timestamp for table names and the like down the line
    now = e.created_at.strftime("%Y%m%d%H%M%S")
    
    # get the integration Columns we'll shop for among StagingTables
    def_cols = Column.objects.filter(definition=d)
    logger.debug("Columns in Definition: %s" % ', '.join([str(x) for x in def_cols]))
    
    # create the empty table
    tbl_clauses = []
    for dcol in d.columns.all():
        # construct a clause representing this column, based on its name and data type
        clause = "{f} {type}".format(f=dcol.field_out.value, type=dcol.field_out_data_type)
        
        # only append the clause if it doesn't already exist in the clauses list
        # (avoids duplicates)
        if clause not in tbl_clauses:
            tbl_clauses.append(clause)
    
    int_tbl_name = 'int_{ds_parent}_{cat}_{now}'.format(
        ds_parent=d.data_source.mr_descriptor,
        cat=d.category,
        now=now,
    )
    logger.info("Creating new database table {tbl} for table integration" \
        .format(tbl=int_tbl_name))
    create = "CREATE TABLE {tbl} ({clauses});" \
        .format(tbl=int_tbl_name, clauses=", ".join(tbl_clauses))
    cursor = cnxn.cursor()
    cursor.execute(create)
    cnxn.commit()
    logger.info("Created new database table {tbl}" \
        .format(tbl=int_tbl_name))
    
    logger.info("Creating new IntegrationTable object for table {tbl}" \
        .format(tbl=int_tbl_name))
    # create a new IntegrationTable to hold results and meta
    int_tbl = IntegrationTable(
        name=int_tbl_name,
        definition=d,
        edition=e
    )
    int_tbl.save()
    logger.info("Created new IntegrationTable object for table {tbl}" \
        .format(tbl=int_tbl_name))
    
    logger.info("Checking for indexes to add to table {tbl}" \
        .format(tbl=int_tbl_name))
    for i in indexes:
        logger.info("Attempting to create index on column {i} in table {tbl}" \
            .format(i=i, tbl=int_tbl_name))
        try:
            #print "i is {i}".format(i=i)
            #print "int_tbl_name is {int_tbl_name}".format(int_tbl_name=int_tbl_name)
            i_sql = pyhcup.db.index_sql(i, int_tbl_name)
            #print i_sql
            cursor.execute(i_sql)
            cnxn.commit()
            logger.info("Created index on column {i} in table {tbl}" \
                .format(i=i, tbl=int_tbl_name))
        except:
            logger.warning("Failed to create index on column {i} in table {tbl}" \
                .format(i=i, tbl=int_tbl_name))
            cnxn.rollback()
    
    # get StagingTables to be included in this integration
    # according to category indicated by Definition
    stgtbls = StagingTable.objects \
                .filter(category=d.category, wip=False, void=False) \
                .order_by('-created_at') # most recent first
    
    # extra filter step to make sure these are all in the same family.
    # this cannot be combined with above since this is a property,
    # rather than a model column
    stgtbls = filter(
        lambda x: x.data_source.parent == d.data_source,
        stgtbls
    )
    
    integrated_sy = []
    shoveled_lst = []
    for sttbl in stgtbls:
        # put a check in here to make sure
        # we do not double-add.
        state = sttbl.state
        year = sttbl.year
        sy = dict(state=state, year=year)
        if sy in integrated_sy:
            # do not proceed
            logger.info("StagingTable {st_pk} ({tbl}) will be omitted, as {st} {yr} have already been dispatched for integration" \
                .format(st_pk=sttbl.pk, st=state, yr=year, tbl=sttbl.name)
            )
        else:
            logger.info("Integrating StagingTable {st_pk} ({tbl})" \
                .format(st_pk=sttbl.pk, tbl=sttbl.name)
            )
            integrated_sy.append(sy)
            dispatched = integrate_stg_table.delay(sttbl.pk, int_tbl.pk, d.pk)
            shoveled_lst.append(dispatched)
    
    callback = integration_complete.delay(shoveled_lst, int_tbl.pk)
    
    return int_tbl


@celery.shared_task(bind=True, max_retries=None, default_retry_delay=900)
def integration_complete(self, shoveled_lst, int_tbl_pk):
    
    int_tbl = IntegrationTable.objects.get(pk=int_tbl_pk)
    d = int_tbl.definition
    
    try:
        # test if all completed, one way or another
        if all(s.ready() for s in shoveled_lst):
            
            # test if the outcomes were all SUCCESS
            if all(s.successful() for s in shoveled_lst):
                
                # log a summary about the entire enterprise
                total_shoveled = sum([x.result['shoveled'] if 'shoveled' in x.result else 0 for x in shoveled_lst])
                logger.info("Integrated a total of {total} rows from {ct} tables" \
                    .format(total=total_shoveled, ct=len(shoveled_lst))
                )
                
                # record some details in the integration table description
                int_tbl.description = 'Integration of {ds} {category} as of {now}\n\nDIGEST\n##########\n\n' \
                    .format(ds=d.data_source, category=d.category, now=datetime.datetime.utcnow().replace(tzinfo=utc))
                int_tbl.description += '\n\n'.join([
                    "State: {st}\nYear: {yr}\nTable in: {tbl}\nRows moved: {rows}\nFields in: {fields_in}\nFields out: {fields_out}" \
                        .format(
                            st=x.result['state'],
                            yr=x.result['year'],
                            tbl=str(x.result['StagingTable']),
                            rows=x.result['shoveled'],
                            fields_in=x.result['fields_in'],
                            fields_out=x.result['fields_out'],
                        )
                    for x in shoveled_lst
                ])
                int_tbl.wip = False
                int_tbl.save()
                return int_tbl
        else:
            # since some haven't completed yet, raise IncompleteStream and retry
            raise self.retry(exc=Exception('Integration not yet complete: %s' % int_tbl))
    
    except celery.exceptions.IncompleteStream as exc:
        # try again a little later, when the streams are more complete
        raise self.retry(exc=Exception('Integration not yet complete: %s' % int_tbl))



# core Python packages
import logging


# third party packages


# django packages
from django.shortcuts import render
from django.utils.timezone import utc
from django.contrib.auth.decorators import login_required


# local imports


# start a logger
logger = logging.getLogger('default')


# Create your views here.
@login_required(login_url='/login/')
def index(request):
    logger.debug('Integration index requested')
    
    context = {
        'title': 'The Django-HCUP Hachoir: Integration Index',
    }
    
    template = 'djhcup_base.html'
    return render(request, template, context)



# Django imports
from django.conf.urls import patterns, include, url


from djhcup_integration.views import index


# base patterns always available through having djhcup_core installed
urlpatterns = patterns('',
    url(r'$', index, name='int_index'),
)


from django.contrib import admin

#import local models
from djhcup_integration.models import Edition, IntegrationTable, Definition, Column, ColumnName, IntTblStgTblMembers, MappedColumn

#Register models
admin.site.register(Edition)
admin.site.register(IntegrationTable)
admin.site.register(Definition)
admin.site.register(Column)
admin.site.register(ColumnName)
admin.site.register(IntTblStgTblMembers)
admin.site.register(MappedColumn)



# core Python packages
import datetime
import logging
import os
import sys
import psycopg2


# third party packages
import pandas as pd
import pyhcup


# django packages
from django.shortcuts import render
from django.http import HttpResponse
from django.utils.timezone import utc
from django.conf import settings
from django.core.urlresolvers import reverse
from django.core.exceptions import ObjectDoesNotExist
from django.db import connections as cnxns
from django.db import transaction as tx


# local project and app packages
from djhcup_integration.models import Edition, Definition, Column, ColumnName, IntegrationTable, IntTblStgTblMembers, MappedColumn


# sister app stuff
from djhcup_staging.models import DataSource, StagingTable, File
from djhcup_staging.utils.misc import dead_batches, get_pgcnxn


# start a logger
logger = logging.getLogger('default')


def master_table(ds_id, category, timestamp=None, extra=None):
    """
    Generates a master table for the DataSource corresponding to ds_id. Returns an IntegrationTable.
    
    Uses all known loadfiles for source and generates one master table.
    
    # deprecated; ran up against max 1600 column limit in PostgreSQL
    # (which is a lot, admittedly).
    # only still in here for historical reasons.
    """
    
    source = DataSource.objects.get(pk=ds_id)
    logger.info('Attempting to generate master table for %s' % source)
    
    cursor = cnxns[DJHCUP_DB_ENTRY].cursor()

    # need later for table creation, but want the value to be based on the start of the process
    if timestamp == None:
        timestamp = datetime.datetime.utcnow().replace(tzinfo=utc).strftime("%Y%m%d%H%M%S")
    
    db_table = "int_%s_%s_%s" % (source.mr_descriptor, category, timestamp)
    db_table = db_table.lower()
    pk_fields = []
    index_fields = []
    
    # check to see whether this source has a LONG_MAPS entry, which would indicate it ought to be a long table
    if source.abbreviation in pyhcup.parser.LONG_MAPS.keys():
        logger.warning("The indicated data source has an analog in pyhcup.parser.LONG_MAPS. Generation will proceed, but this may indicate an error.")
    
    src_loadfiles = File.objects.filter(
        file_type__content_type__in=['LOAD'],
        file_type__source=source
    )
    logger.info('Found %d loadfile records matching source' % len(src_loadfiles))
    
    meta_list = []
    excluded = []
    for lf in src_loadfiles:
        mdf = pyhcup.sas.meta_from_sas(lf.full_path)
        augmented = pyhcup.meta.augment(mdf)
        meta_list.append(augmented)
    
    merged_meta = pyhcup.meta.merge(meta_list)
    logger.info('Merged %d meta DataFrames successfully, with %d columns after de-duplication' % (len(meta_list), len(merged_meta)))
    
    constraints = ['NULL']
    pk_fields = ['year', 'state', 'key']
    index_fields += pk_fields
    
    create_sql = pyhcup.db.table_sql(
        merged_meta, db_table, pk_fields=pk_fields, ine=True,
        append_state=True, default_constraints=constraints
    )
        
    try:
        cursor.execute(create_sql)
        tx.commit(using=DJHCUP_DB_ENTRY)
    except:
        e = sys.exc_info()[0]
        raise Exception("Failed to create master table with query %s (%s)" % (create_sql, e))
    
    logger.info('Created master table %s with on database %s' % (db_table, DJHCUP_DB_ENTRY))
    
    if category in EXTRA_COLUMNS.keys():
        # clean out any columns that will be added back on
        # re-add them with proper definitions
        for c in EXTRA_COLUMNS[category]:
            dropif_sql = 'ALTER TABLE {t} DROP COLUMN IF EXISTS {c};' \
                .format(t=db_table, c=c['field'])
            add_sql = 'ALTER TABLE {t} ADD COLUMN {c} {coltype};' \
                .format(t=db_table, c=c['field'], coltype=c['type'])
            logger.info('Dropping column if exists: {c}' \
                .format(c=c['field']))
            cursor.execute(dropif_sql)
            logger.info('Adding column: {c} {coltype}' \
                .format(c=c['field'], coltype=c['type']))
            cursor.execute(add_sql)
        
        tx.commit(using=DJHCUP_DB_ENTRY)
    
    if index_fields != None:
        try:
            for col in index_fields:
                index = pyhcup.db.index_sql(col, db_table)
                cursor.execute(index)
                tx.commit(using=DJHCUP_DB_ENTRY)
                logger.info('Created index for column %s with on table %s' % (col, db_table))
        except:
            raise Exception("Failed to create indexes on table, particularly one that should have resulted from this query: %s" % index)
    
    table_record = IntegrationTable(
        data_source=source,
        name=db_table,
        database=DJHCUP_DB_ENTRY,
        category=category
    )
    table_record.save()
    logger.info('Saved master table details as IntegrationTable record %d' % table_record.pk)
    
    return table_record



# core Python packages
import logging


# third party packages


# django packages
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned


# local project and app packages
from djhcup_integration.models import ColumnName, Column, Definition
from djhcup_staging.models import DataSource

# move these into the local namespace
from djhcup_integration.config import DEFAULT_DEFINITIONS

# kludge workaround--move to config later
DJHCUP_DB_ENTRY = 'djhcup'

# start a logger
logger = logging.getLogger('default')


def populate():
    definitions = []
    colnames = []
    cols = []
    
    for d in DEFAULT_DEFINITIONS:
        try:
            name = d['name']
            category = d['category']
            abbr = d['ds_family_abbr']
            p_abbr = d['ds_parent_abbr']
            
            try:
                # isolate the DataSource
                ds = DataSource.objects.get(
                    abbreviation=abbr,
                    parent__abbreviation=p_abbr
                )
                
                try:
                    def_obj = Definition.objects.get(
                        data_source=ds,
                        category=category,
                        name=name
                    )
                    logger.info("Found existing Definition object {pk}" \
                        .format(pk=def_obj.pk))
                    
                except (ObjectDoesNotExist, MultipleObjectsReturned):
                    def_obj = Definition(
                        data_source=ds,
                        category=category,
                        name=name,
                    )
                    def_obj.save()
                    logger.info("Added Definition object {pk} for {name}" \
                        .format(pk=def_obj.pk, name=def_obj.name))
                
                for i in d['columns']:
                    # move on to column logging
                    try:
                        cn = ColumnName.objects.get(value__iexact=i['field'])
                    except:
                        cn = ColumnName(value=i['field'])
                        cn.save()
                        colnames.append(cn)
                    
                    try:
                        col = Column.objects.get(fields_in=cn, field_out__value__iexact=i['field'])
                    except:
                        col = Column(field_out=cn, field_out_data_type=i['type'])
                        col.save()
                        col.fields_in.add(cn)
                        cols.append(col)
                    
                    def_obj.columns.add(col)
                
                logger.info("Added or verified {ct} Column objects for Definition object {pk}" \
                    .format(ct=str(def_obj.columns.all().count()), pk=def_obj.pk))
                
                definitions.append(def_obj)
            
            except (ObjectDoesNotExist, MultipleObjectsReturned):
                logger.warning(
                    "Unable to isolate djhcup_staging.models.DataSource " +\
                    "object with abbreviation {a} and parent " +\
                    "abbreviation {p}. Definition for {d} will not " +\
                    "proceed".format(a=abbr, p=p_abbr, d=name)
                )
        except KeyError:
            logger.error(
                "Definition lacks critical components. Will not " +\
                "proceed.\n\n{d}".format(d=d)
            )
    
    return definitions, colnames, cols



import psycopg2
import logging
import pyhcup

from django.conf import settings
from django.db import models


from djhcup_staging.models import StagingTable, DataSource


# pull these into local ns for psycopg2/dbapi2 objects
DB_DEF = settings.DATABASES['djhcup']

# start a logger
default_logger = logging.getLogger('default')


class Edition(models.Model):
    name = models.CharField(max_length=250, blank=False)
    description = models.TextField(blank=True, null=True, default=None)
    created_at = models.DateTimeField(auto_now_add=True)
    void = models.BooleanField(default=False)
    voided_at = models.DateTimeField(blank=True, null=True, default=None)
    
    def __unicode__(self):
        return "<Edition {pk}>".format(pk=self.pk)
    
    def __repr__(self):
        return self.__unicode__()


class IntegrationTable(models.Model):
    """This is a record of integration tables created/populated.
    
    stagingtables should share category and data_source
    
    IntegrationTables should always have state and year fields.
    """
    stagingtables = models.ManyToManyField(
        StagingTable,
        help_text="StagingTable objects comprising this IntegrationTable",
        through = 'IntTblStgTblMembers'
        )
    name = models.CharField(max_length=250, blank=False)
    description = models.TextField(blank=True, null=True, default=None)
    edition = models.ForeignKey('Edition')
    definition = models.ForeignKey('Definition', blank=True, null=True)
    schema = models.CharField(max_length=250, blank=True)
    database = models.CharField(max_length=250, blank=True, help_text="Leave blank to use default connection specified in Django's setings.py file")
    wip = models.BooleanField(
        default=True,
        help_text="This table is work-in-progress, and should not be used."
    )
    
    #meta cruft
    created_at = models.DateTimeField(auto_now_add=True)
    void = models.BooleanField(default=False)
    voided_at = models.DateTimeField(blank=True, null=True, default=None)
    records = models.IntegerField(blank=True, null=True, default=None)
    
    # overwrite to drop associated table by default when deleting
    def delete(self, drop_table=True, *args, **kwargs):
        if drop_table:
            cnxn = psycopg2.connect(
                host=DB_DEF['HOST'],
                port=DB_DEF['PORT'],
                user=DB_DEF['USER'],
                password=DB_DEF['PASSWORD'],
                database=DB_DEF['NAME'],
            )
            drop_result = pyhcup.db.pg_drop(cnxn, self.name)
            if drop_result == True:
                default_logger.info('Dropped table {table} from database'.format(table=self.name))
            else:
                default_logger.error('Unable to drop table {table} from database'.format(table=self.name))
        
        super(IntegrationTable, self).delete(*args, **kwargs)
    
    
    def __unicode__(self):
        try:
            d = self.definition.pk
        except:
            d = None
        
        try:
            e = self.edition.pk
        except:
            e = None
        
        return "<IntegrationTable {pk}: {n}, definition={d} edition={e}>".format(pk=self.pk, n=self.name, d=d, e=e)
    
    def __repr__(self):
        return self.__unicode__()
    
    """
    def get_absolute_url(self):
        return reverse(
            'djhcup_integration.views.obj_detail',
            kwargs={'obj_id': self.pk, 'obj_type': 'IntegrationTable'}
        )
    """


class Definition(models.Model):
    """
    Groups a set of integration Column objects. More or less.
    
    Subsequently used to create an IntegrationTable.
    """
    
    data_source = models.ForeignKey(
        DataSource,
        verbose_name="DataSource Family",
        help_text="""
            DataSource object from djhcup_staging. Restricts integration to tables
            with this DataSource as parent to their own; e.g., HCUP SID.
            """,
        blank=True, # this is optional,
        null=True
    )
    TBL_CAT_CHOICES = [
        ('CHARGES','Charges data in long format (converted from wide source if necessary)'),
        ('DX','Diagnoses data in long format (from CORE_SOURCE)'),
        ('PR','Procedures data in long format (from CORE_SOURCE)'),
        ('UFLAGS','Utilization flag data, long format (from CORE_SOURCE, CHARGES, and PR)'),
        ('CORE_PROCESSED','CORE Processed: CORE, DaysToEvent/VisitLink, wide charges, wide uflags'),
    ]
    
    category = models.CharField(
        max_length=100,
        choices=TBL_CAT_CHOICES,
        blank=False
    )
    name = models.CharField(max_length=100, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    columns = models.ManyToManyField('Column')
    # record index columns here? as extra info on M2M?
    
    def __unicode__(self):
        return '<Integration Definition: {n} ({c})>' \
            .format(n=self.name, c=self.category)
    
    def __repr__(self):
        return self.__unicode__()


class IntTblStgTblMembers(models.Model):
    integrationtable = models.ForeignKey('IntegrationTable')
    stagingtable = models.ForeignKey(StagingTable)
    mappedcols = models.ManyToManyField('Column', through='MappedColumn')
    unmappedcols = models.ManyToManyField('Column', related_name='unmappedcols')
    
    def __unicode__(self):
        return '<IntTblStgTblMember: StgTbl {s} in IntTbl {i}>' \
            .format(s=self.stagingtable.pk, i=self.integrationtable.pk)
    
    def __repr__(self):
        return self.__unicode__()


class MappedColumn(models.Model):
    inttblstgtblmembers = models.ForeignKey('IntTblStgTblMembers')
    column = models.ForeignKey('Column')
    matched_in = models.ForeignKey('ColumnName')


class Column(models.Model):
    """
    Personally I believe this would be better if fields_in actually referred
    to djhcup_staging.models.Column objects, but then I'd need another layer
    that establishes the list of column names which qualify for inclusion.
    Reason being that staging columns are not propagated as part of table
    processing--they exist only as part of import loading.
    
    
    Might still be good to do at a future date but right now is kind of a
    hassle.
    """
    fields_in = models.ManyToManyField('ColumnName', related_name='fields_in')
    field_out = models.ForeignKey('ColumnName', related_name='field_out')
    TYPE_CHOICES = [
            ('INT', 'Integer'),
            ('BIGINT', 'Very Large Integer'),
            ('TEXT', 'Text field'),
            ('VARCHAR', 'Character varying'),
            ('NUMERIC(2, 15)', 'Decimal, out to hundredths'),
        ]
    field_out_data_type = models.CharField(max_length=15, choices=TYPE_CHOICES, default='VARCHAR')
    field_out_scale = models.IntegerField(blank=True, null=True, default=None)
    field_out_precision = models.IntegerField(blank=True, null=True, default=None)
    
    def _is_numeric(self):
        lower_type = self.field_out_data_type.lower()
        if (lower_type in ['int', 'bigint'] or
            'numeric' in lower_type):
            return True
        else:
            return False
    
    def _name(self):
        return self.field_out.value
    name = property(_name)
    
    def __unicode__(self):
        return '<Integration Column: {i} to {o} ({pk})>' \
            .format(i=[str(x.value) for x in self.fields_in.all()],
                    o=self.field_out.value, pk=self.pk)


class ColumnName(models.Model):
    value = models.CharField(max_length=150, blank=False)
    
    def __unicode__(self):
        return 'ColumnName: {v}'.format(v=self.value)
